@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	1 August 1988
- * Last Edited:	29 August 1996
+ * Last Edited:	10 December 1998
  *
- * Copyright 1996 by the University of Washington
+ * Copyright 1998 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -36,15 +36,30 @@
 
 static char *myUserName = NIL;	/* user name */
 static char *myLocalHost = NIL;	/* local host name */
+static char *myClientHost = NIL;/* client host name */
+static char *myServerHost = NIL;/* server host name */
 static char *myHomeDir = NIL;	/* home directory name */
 static char *myNewsrc = NIL;	/* newsrc file name */
 static char *sysInbox = NIL;	/* system inbox name */
 static long list_max_level = 5;	/* maximum level of list recursion */
+				/* home namespace */
+static NAMESPACE nshome = {"",'\\',NIL,NIL};
+				/* namespace list */
+static NAMESPACE *nslist[3] = {&nshome,NIL,NIL};
 static long alarm_countdown = 0;/* alarm count down */
 static void (*alarm_rang) ();	/* alarm interrupt function */
 static unsigned int rndm = 0;	/* initial `random' number */
+static int logtry = 3;		/* number of login tries */
+				/* callback to get username */
+static userprompt_t mailusername = NIL;
+static long is_nt = -1;		/* T if NT, NIL if not NT, -1 unknown */
 
 #include "write.c"		/* include safe writing routines */
+
+
+/* Get all authenticators */
+
+#include "auths.c"
 
 /* Environment manipulate parameters
  * Accepts: function code
@@ -55,7 +70,19 @@ static unsigned int rndm = 0;	/* initial `random' number */
 void *env_parameters (long function,void *value)
 {
   switch ((int) function) {
+  case SET_NAMESPACE:
+    fatal ("SET_NAMESPACE not permitted");
+  case GET_NAMESPACE:
+    value = (void *) nslist;
+    break;
+  case SET_USERPROMPT :
+    mailusername = (userprompt_t) value;
+    break;
+  case GET_USERPROMPT :
+    value = (void *) mailusername;
+    break;
   case SET_HOMEDIR:
+    if (myHomeDir) fs_give ((void **) &myHomeDir);
     myHomeDir = cpystr ((char *) value);
     break;
   case GET_HOMEDIR:
@@ -65,6 +92,7 @@ void *env_parameters (long function,void *value)
     myLocalHost = cpystr ((char *) value);
     break;
   case GET_LOCALHOST:
+    if (myLocalHost) fs_give ((void **) &myLocalHost);
     value = (void *) myLocalHost;
     break;
   case SET_NEWSRC:
@@ -73,7 +101,6 @@ void *env_parameters (long function,void *value)
     break;
   case GET_NEWSRC:
     if (!myNewsrc) {		/* set news file name if not defined */
-
       char tmp[MAILTMPLEN];
       sprintf (tmp,"%s\\NEWSRC",myhomedir ());
       myNewsrc = cpystr (tmp);
@@ -163,7 +190,7 @@ void internal_date (char *date)
 /* Return random number
  */
 
-long random ()
+long random (void)
 {
   if (!rndm) srand (rndm = (unsigned) time (0L));
   return (long) rand ();
@@ -188,51 +215,176 @@ void CALLBACK clock_ticked (UINT IDEvent,UINT uReserved,DWORD dwUser,
 {
   if (alarm_rang && !--alarm_countdown) (*alarm_rang) ();
 }
-
-
-/* Set server traps
- * Accepts: clock interrupt handler
+
+/* Initialize server
+ * Accepts: server name for syslog or NIL
+ *	    /etc/services service name or NIL
+ *	    alternate /etc/services service name or NIL
+ *	    SASL service name or NIL
+ *	    clock interrupt handler
  *	    kiss-of-death interrupt handler
+ *	    hangup interrupt handler
+ *	    termination interrupt handler
  */
 
-void server_traps (void *clkint,void *kodint)
+void server_init (char *server,char *service,char *altservice,char *sasl,
+		  void *clkint,void *kodint,void *hupint,void *trmint)
 {
+  if (server) {			/* set server name in syslog */
+    openlog (server,LOG_PID,LOG_MAIL);
+    fclose (stderr);		/* possibly save a process ID */
+  }
+				/* set SASL name */
+  if (sasl) mail_parameters (NIL,SET_SERVICENAME,(void *) sasl);
   alarm_rang = clkint;		/* note the clock interrupt */
   timeBeginPeriod (1000);	/* set the timer interval */
   timeSetEvent (1000,1000,clock_ticked,NIL,TIME_PERIODIC);
 }
-
-/* Log in as anonymous daemon
- * Returns: T if successful, NIL if error
+
+
+/* Wait for stdin input
+ * Accepts: timeout in seconds
+ * Returns: T if have input on stdin, else NIL
  */
 
-long anonymous_login ()
+long server_input_wait (long seconds)
 {
-  return server_login ("Guest",NIL,NIL,NIL);
+  fd_set rfd;
+  struct timeval tmo;
+  FD_ZERO (&rfd);
+  FD_SET (0,&rfd);
+  tmo.tv_sec = seconds; tmo.tv_usec = 0;
+  return select (1,&rfd,0,0,&tmo) ? LONGT : NIL;
+}
+
+/* Server log in
+ * Accepts: user name string
+ *	    password string
+ *	    argument count
+ *	    argument vector
+ * Returns: T if password validated, NIL otherwise
+ */
+
+static int gotprivs = NIL;	/* once-only flag to grab privileges */
+
+long server_login (char *user,char *pass,int argc,char *argv[])
+{
+  HANDLE hdl;
+  LUID tcbpriv;
+  TOKEN_PRIVILEGES tkp;
+  char *s;
+  if (!gotprivs++) {		/* need to get privileges? */
+				/* yes, note client host if specified */
+    if (argc == 2) myClientHost = argv[1];
+				/* get process token and TCB priv value */
+    if (!(OpenProcessToken (GetCurrentProcess (),
+			    TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,&hdl) &&
+	  LookupPrivilegeValue ((LPSTR) NIL,SE_TCB_NAME,&tcbpriv)))
+      return NIL;
+    tkp.PrivilegeCount = 1;	/* want to enable this privilege */
+    tkp.Privileges[0].Luid = tcbpriv;
+    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
+				/* enable it */
+    AdjustTokenPrivileges (hdl,NIL,&tkp,sizeof (TOKEN_PRIVILEGES),
+			   (PTOKEN_PRIVILEGES) NIL,(PDWORD) NIL);
+				/* make sure it won */
+    if (GetLastError() != ERROR_SUCCESS) return NIL;
+  }
+				/* cretins still haven't given up */
+  if (strlen (user) >= MAILTMPLEN)
+    syslog (LOG_ALERT,"System break-in attempt, host=%.80s",tcp_clienthost ());
+  else if ((logtry > 0) &&	/* try to login and impersonate the guy */
+	   ((LogonUser (user,".",pass,LOGON32_LOGON_INTERACTIVE,
+			LOGON32_PROVIDER_DEFAULT,&hdl) ||
+	     LogonUser (user,".",pass,LOGON32_LOGON_BATCH,
+			LOGON32_PROVIDER_DEFAULT,&hdl) ||
+	     LogonUser (user,".",pass,LOGON32_LOGON_SERVICE,
+			LOGON32_PROVIDER_DEFAULT,&hdl)) &&
+	    ImpersonateLoggedOnUser (hdl))) return env_init (user,NIL);
+  s = (logtry-- > 0) ? "Login failure" : "Excessive login attempts";
+				/* note the failure in the syslog */
+  syslog (LOG_INFO,"%s user=%.80s host=%.80s",s,user,tcp_clienthost ());
+  sleep (3);			/* slow down possible cracker */
+  return NIL;
+}
+
+/* Authenticated server log in
+ * Accepts: user name string
+ *	    argument count
+ *	    argument vector
+ * Returns: T if password validated, NIL otherwise
+ */
+
+long authserver_login (char *user,int argc,char *argv[])
+{
+  return server_login (user,NIL,argc,argv);
 }
 
 
+/* Log in as anonymous daemon
+ * Accepts: argument count
+ *	    argument vector
+ * Returns: T if successful, NIL if error
+ */
+
+long anonymous_login (int argc,char *argv[])
+{
+  return server_login ("Guest",NIL,argc,argv);
+}
+
 /* Initialize environment
- * Accepts: user name
- *	    home directory or NIL to use value from user profile
+ * Accepts: user name, or NIL to skip setting user name on (Win9x only)
+ *          home directory, or NIL to use user info (NT only)
  * Returns: T, always
  */
 
+typedef NET_API_STATUS (CALLBACK *NUGI)
+     (LPCWSTR servername,LPCWSTR username,DWORD level,LPBYTE *bufptr);
+
+
 long env_init (char *user,char *home)
 {
-  char tmp[MAILTMPLEN];
-  PUSER_INFO_1 ui;
   if (myUserName) fatal ("env_init called twice!");
-  myUserName = cpystr (user);	/* remember user name and home directory */
-  myHomeDir = cpystr ((home && *home) ? home :
-		      ((MultiByteToWideChar (CP_ACP,0,user,strlen (user)+1,
-					    (WCHAR *) tmp,MAILTMPLEN) &&
-			!NetUserGetInfo (NIL,(LPWSTR) &tmp,1,(LPBYTE *) &ui) &&
-			WideCharToMultiByte (CP_ACP,0,ui->usri1_home_dir,-1,
-					     tmp,MAILTMPLEN,NIL,NIL) &&
-			tmp[0]) ? tmp : defaultPath (tmp)));
+				/* remember user name */
+  if (user && *user) myUserName = cpystr (user);
+  else if (check_nt ()) fatal ("missing user name argument to env_init!");
+  if (!myHomeDir) {		/* only if home directory not set up yet */
+				/* remember home directory */
+    if (home && *home) myHomeDir = cpystr (home);
+    else if (check_nt ()) {	/* get from user info on NT */
+      char tmp[MAILTMPLEN];
+      PUSER_INFO_1 ui;
+      NUGI nugi;
+      HINSTANCE nahdl = LoadLibrary ("netapi32.dll");
+      if (!(nahdl && (nugi = (NUGI) GetProcAddress (nahdl,"NetUserGetInfo")) &&
+	    MultiByteToWideChar (CP_ACP,0,user,strlen (user) + 1,
+				 (WCHAR *) tmp,MAILTMPLEN) &&
+	    !(*nugi) (NIL,(LPWSTR) &tmp,1,(LPBYTE *) &ui) &&
+	    WideCharToMultiByte (CP_ACP,0,ui->usri1_home_dir,-1,
+				 tmp,MAILTMPLEN,NIL,NIL) && tmp[0]))
+	sprintf (tmp,"%s%s",defaultDrive (),"\\users\\default");
+      myHomeDir = cpystr (tmp);	/* remember home directory */
+    }
+    else fatal ("missing home directory argument to env_init!");
+  }
   return T;
 }
+
+
+/* Check if NT
+ * Returns: T if NT, NIL if Win9x
+ */
+
+int check_nt (void)
+{
+  if (is_nt < 0) {		/* not yet set up? */
+    OSVERSIONINFO ver;
+    ver.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
+    GetVersionEx (&ver);
+    is_nt = (ver.dwPlatformId == VER_PLATFORM_WIN32_NT) ? T : NIL;
+  }
+  return is_nt;
+}
 
 /* Return default drive
  * Returns: default drive
@@ -245,41 +397,24 @@ static char *defaultDrive (void)
 }
 
 
-/* Return default path
- * Accepts: path to write into
- * Returns: default path
- */
-
-static char *defaultPath (char *path)
-{
-  sprintf (path,"%s%s",defaultDrive (),"\\users\\default");
-  return path;
-}
-
-
-/* Return home drive from environment variables
- * Returns: home drive
- */
-
-static char *homeDrive (void)
-{
-  char *s;
-  return ((s = getenv ("HOMEDRIVE")) && *s) ? s : defaultDrive ();
-}
-
-
 /* Return home path from environment variables
  * Accepts: path to write into
- * Returns: home path or NIL if it can't be determined
+ * Returns: home path, or NIL if it can't be determined (NT only)
  */
 
 static char *homePath (char *path)
 {
   int i;
   char *s;
-  if (!((s = getenv ("HOMEPATH")) && (i = strlen (s)))) return NIL;
-  if (((s[i-1] == '\\') || (s[i-1] == '/'))) s[i-1] = '\0';
-  sprintf (path,"%s%s",homeDrive (),s);
+				/* get home drive */
+  strcpy (path,((s = getenv ("HOMEDRIVE")) && *s) ? s : defaultDrive ());
+				/* if have a home path */
+  if ((s = getenv ("HOMEPATH")) && (i = strlen (s))) {
+    if (((s[i-1] == '\\') || (s[i-1] == '/'))) s[i-1] = '\0';
+    strcat (path,s);		/* append the home path to the drive */
+  }
+				/* return NIL if NT and no home path */
+  else if (check_nt ()) return NIL;
   return path;
 }
 
@@ -290,20 +425,23 @@ static char *homePath (char *path)
 
 char *myusername_full (unsigned long *flags)
 {
+  HANDLE tok;
+  UCHAR tmp[MAILTMPLEN],user[MAILTMPLEN],domain[MAILTMPLEN];
+  DWORD len,userlen = MAILTMPLEN,domainlen = MAILTMPLEN;
+  SID_NAME_USE snu;
   char *ret = "SYSTEM";
   if (!myUserName) {		/* get user name if don't have it yet */
-    HANDLE tok;
-    UCHAR tmp[MAILTMPLEN],user[MAILTMPLEN],domain[MAILTMPLEN];
-    DWORD len,userlen = MAILTMPLEN,domainlen = MAILTMPLEN;
-    SID_NAME_USE snu;
- 				/* lookup user and domain */
-    if (!(OpenProcessToken (GetCurrentProcess (),TOKEN_READ,&tok) &&
-	  GetTokenInformation (tok,TokenUser,tmp,MAILTMPLEN,&len) &&
-	  LookupAccountSid (NIL,((PTOKEN_USER) tmp)->User.Sid,user,
-			    &userlen,domain,&domainlen,&snu)))
-      fatal ("Unable to look up user name");
+    if (check_nt ()) {		/* NT lookup user and domain */
+      if (!(OpenProcessToken (GetCurrentProcess (),TOKEN_READ,&tok) &&
+	    GetTokenInformation (tok,TokenUser,tmp,MAILTMPLEN,&len) &&
+	    LookupAccountSid (NIL,((PTOKEN_USER) tmp)->User.Sid,user,
+			      &userlen,domain,&domainlen,&snu)))
+	fatal ("Unable to look up user name");
 				/* init environment if not SYSTEM */
-    if (_stricmp (user,"SYSTEM")) env_init (user,homePath (tmp));
+      if (_stricmp (user,"SYSTEM")) env_init (user,homePath (tmp));
+    }
+				/* Win9x external username */
+    else env_init (mailusername ? (*mailusername) () : NIL,homePath (tmp));
   }
   if (myUserName) {		/* logged in? */
     if (flags)			/* Guest is an anonymous user */
@@ -322,7 +460,7 @@ char *myusername_full (unsigned long *flags)
 char *myhomedir ()
 {
   if (!myHomeDir) myusername ();/* initialize if first time */
-  return myHomeDir ? myHomeDir : homeDrive ();
+  return myHomeDir ? myHomeDir : "";
 }
 
 /* Return system standard INBOX
@@ -333,7 +471,8 @@ char *sysinbox ()
 {
   char tmp[MAILTMPLEN];
   if (!sysInbox) {		/* initialize if first time */
-    sprintf (tmp,MAILFILE,myUserName);
+    if (check_nt ()) sprintf (tmp,MAILFILE,myUserName);
+    else sprintf (tmp,"%s\\INBOX",myhomedir ());
     sysInbox = cpystr (tmp);	/* system inbox is from mail spool */
   }
   return sysInbox;
@@ -376,11 +515,10 @@ int lockname (char *lock,char *fname,int op)
 {
   int ld;
   char c,*s;
-  if (((s = getenv ("TEMP")) && *s) || ((s = getenv ("TEMPDIR")) && *s)) {
+  if (((s = getenv ("TEMP")) && *s) || ((s = getenv ("TMPDIR")) && *s))
     strcpy (lock,s);
-    if ((s = lock + strlen (lock))[-1] != '\\') *s++ ='\\';
-  }
   else sprintf (lock,"%s\\TEMP\\",defaultDrive ());
+  if ((s = lock + strlen (lock))[-1] != '\\') *s++ ='\\';
   while (c = *fname++) switch (c) {
   case '/':
   case '\\':
@@ -406,85 +544,18 @@ int lockname (char *lock,char *fname,int op)
 
 void unlockfd (int fd,char *lock)
 {
-  unlink (lock);		/* delete the file */
   flock (fd,LOCK_UN);		/* unlock it */
   close (fd);			/* close it */
 }
 
 
 /* Determine default prototype stream to user
+ * Accepts: type (NIL for create, T for append)
  * Returns: default prototype stream
  */
 
-MAILSTREAM *default_proto ()
+MAILSTREAM *default_proto (long type)
 {
   extern MAILSTREAM DEFAULTPROTO;
   return &DEFAULTPROTO;		/* return default driver's prototype */
 }
-
-/* Copy Unix string with CRLF newlines
- * Accepts: destination string
- *	    pointer to size of destination string buffer
- *	    source string
- *	    length of source string
- * Returns: length of copied string
- */
-
-unsigned long unix_crlfcpy (char **dst,unsigned long *dstl,char *src,
-			    unsigned long srcl)
-{
-  unsigned long i,j;
-  char *d = src;
-				/* count number of LF's in source string(s) */
-  for (i = srcl,j = 0; j < srcl; j++) if (*d++ == '\012') i++;
-				/* flush destination buffer if too small */
-  if (*dst && (i > *dstl)) fs_give ((void **) dst);
-  if (!*dst) {			/* make a new buffer if needed */
-    *dst = (char *) fs_get ((*dstl = i) + 1);
-    if (dstl) *dstl = i;	/* return new buffer length to main program */
-  }
-  d = *dst;			/* destination string */
-				/* copy strings, inserting CR's before LF's */
-  while (srcl--) switch (*src) {
-  case '\015':			/* unlikely carriage return */
-    *d++ = *src++;		/* copy it and any succeeding linefeed */
-    if (srcl && *src == '\012') {
-      *d++ = *src++;
-      srcl--;
-    }
-    break;
-  case '\012':			/* line feed? */
-    *d++ ='\015';		/* yes, prepend a CR, drop into default case */
-  default:			/* ordinary chararacter */
-    *d++ = *src++;		/* just copy character */
-    break;
-  }
-  *d = '\0';			/* tie off destination */
-  return d - *dst;		/* return length */
-}
-
-/* Length of Unix string after unix_crlfcpy applied
- * Accepts: source string
- * Returns: length of string
- */
-
-unsigned long unix_crlflen (STRING *s)
-{
-  unsigned long pos = GETPOS (s);
-  unsigned long i = SIZE (s);
-  unsigned long j = i;
-  while (j--) switch (SNX (s)) {/* search for newlines */
-  case '\015':			/* unlikely carriage return */
-    if (j && (CHR (s) == '\012')) {
-      SNX (s);			/* eat the line feed */
-      j--;
-    }
-    break;
-  case '\012':			/* line feed? */
-    i++;
-  default:			/* ordinary chararacter */
-    break;
-  }
-  SETPOS (s,pos);		/* restore old position */
-  return i;
-}
