@@ -10,9 +10,9 @@
  *		Internet: MRC@CAC.Washington.EDU
  *
  * Date:	20 December 1989
- * Last Edited:	19 June 1998
+ * Last Edited:	22 January 1999
  *
- * Copyright 1998 by the University of Washington
+ * Copyright 1999 by the University of Washington
  *
  *  Permission to use, copy, modify, and distribute this software and its
  * documentation for any purpose and without fee is hereby granted, provided
@@ -139,7 +139,8 @@ long unix_isvalid (char *name,char *tmp)
   struct utimbuf times;
   errno = EINVAL;		/* assume invalid argument */
 				/* must be non-empty file */
-  if ((t = dummy_file (file,name)) && !stat (t,&sbuf)) {
+  if ((t = dummy_file (file,name)) && !stat (t,&sbuf) &&
+      ((sbuf.st_mode & S_IFMT) == S_IFREG)) {
     if (!sbuf.st_size)errno = 0;/* empty file */
     else if ((fd = open (file,O_BINARY|O_RDONLY,NIL)) >= 0) {
 				/* error -1 for invalid format */
@@ -151,6 +152,12 @@ long unix_isvalid (char *name,char *tmp)
       utime (file,&times);	/* set the times */
     }
   }
+				/* in case INBOX but not unix format */
+  else if ((errno == ENOENT) && ((name[0] == 'I') || (name[0] == 'i')) &&
+	   ((name[1] == 'N') || (name[1] == 'n')) &&
+	   ((name[2] == 'B') || (name[2] == 'b')) &&
+	   ((name[3] == 'O') || (name[3] == 'o')) &&
+	   ((name[4] == 'X') || (name[4] == 'x')) && !name[5]) errno = -1;
   return ret;			/* return what we should */
 }
 
@@ -247,7 +254,7 @@ long unix_create (MAILSTREAM *stream,char *mailbox)
   int fd;
   time_t ti = time (0);
   if (!(s = dummy_file (mbx,mailbox))) {
-    sprintf (tmp,"Can't create %s: invalid name",mailbox);
+    sprintf (tmp,"Can't create %.80s: invalid name",mailbox);
     mm_log (tmp,ERROR);
   }
 				/* create underlying file */
@@ -256,7 +263,7 @@ long unix_create (MAILSTREAM *stream,char *mailbox)
     if ((s = strrchr (s,'\\')) && !s[1]) ret = T;
     else if ((fd = open (mbx,O_WRONLY,
 		    (int) mail_parameters (NIL,GET_MBXPROTECTION,NIL))) < 0) {
-      sprintf (tmp,"Can't reopen mailbox node %s: %s",mbx,strerror (errno));
+      sprintf (tmp,"Can't reopen mailbox node %.80s: %s",mbx,strerror (errno));
       mm_log (tmp,ERROR);
       unlink (mbx);		/* delete the file */
     }
@@ -269,7 +276,7 @@ long unix_create (MAILSTREAM *stream,char *mailbox)
 	       pseudo_name,pseudo_from,mylocalhost (),pseudo_subject,ti,
 	       pseudo_msg);
       if ((write (fd,tmp,strlen (tmp)) < 0) || close (fd)) {
-	sprintf (tmp,"Can't initialize mailbox node %s: %s",mbx,
+	sprintf (tmp,"Can't initialize mailbox node %.80s: %s",mbx,
 		 strerror (errno));
 	mm_log (tmp,ERROR);
 	unlink (mbx);		/* delete the file */
@@ -302,32 +309,50 @@ long unix_delete (MAILSTREAM *stream,char *mailbox)
 long unix_rename (MAILSTREAM *stream,char *old,char *newname)
 {
   long ret = NIL;
-  char *s,tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN],lockx[MAILTMPLEN];
+  char c,*s;
+  char tmp[MAILTMPLEN],file[MAILTMPLEN],lock[MAILTMPLEN],lockx[MAILTMPLEN];
   int fd,ld;
+  struct stat sbuf;
   mm_critical (stream);		/* get the c-client lock */
   if (newname && !((s = dummy_file (tmp,newname)) && *s))
-    sprintf (tmp,"Can't rename mailbox %s to %s: invalid name",old,newname);
+    sprintf (tmp,"Can't rename mailbox %.80s to %.80s: invalid name",
+	     old,newname);
   else if ((ld = lockname (lock,dummy_file (file,old),NIL)) < 0)
-    sprintf (tmp,"Can't get lock for mailbox %s",old);
+    sprintf (tmp,"Can't get lock for mailbox %.80s",old);
   else {			/* lock out other c-clients */
     if (flock (ld,LOCK_EX|LOCK_NB)) {
       close (ld);		/* couldn't lock, give up on it then */
-      sprintf (tmp,"Mailbox %s is in use by another process",old);
+      sprintf (tmp,"Mailbox %.80s is in use by another process",old);
     }
 				/* lock out non c-client applications */
     else if ((fd = unix_lock (file,O_BINARY|O_RDWR,S_IREAD|S_IWRITE,lockx,
 			      LOCK_EX)) < 0)
-      sprintf (tmp,"Can't lock mailbox %s: %s",old,strerror (errno));
+      sprintf (tmp,"Can't lock mailbox %.80s: %s",old,strerror (errno));
     else {
       unix_unlock(fd,NIL,lockx);/* pacify evil NTFS */
       if (newname) {		/* want rename? */
-	if (rename (file,s))	/* rename the file */
-	  sprintf (tmp,"Can't rename mailbox %s to %s: %s",old,newname,
+				/* found superior to destination name? */
+	if (s = strrchr (s,'\\')) {
+	  c = *++s;		/* remember first character of inferior */
+	  *s = '\0';		/* tie off to get just superior */
+				/* name doesn't exist, create it */
+	  if ((stat (tmp,&sbuf) || ((sbuf.st_mode & S_IFMT) != S_IFDIR)) &&
+	      !dummy_create (stream,tmp)) {
+	    flock (ld,LOCK_UN);	/* release c-client lock */
+	    close (ld);		/* close c-client lock */
+	    unlink (lock);	/* and delete it */
+	    mm_nocritical (stream);
+	    return ret;		/* return success or failure */
+	  }
+	  *s = c;		/* restore full name */
+	}
+	if (rename (file,tmp))
+	  sprintf (tmp,"Can't rename mailbox %.80s to %.80s: %s",old,newname,
 		   strerror (errno));
 	else ret = T;		/* set success */
       }
       else if (unlink (file))	/* want delete */
-	sprintf (tmp,"Can't delete mailbox %s: %s",old,strerror (errno));
+	sprintf (tmp,"Can't delete mailbox %.80s: %s",old,strerror (errno));
       else ret = T;		/* set success */
       flock (ld,LOCK_UN);	/* release c-client lock */
       close (ld);		/* close c-client lock */
@@ -346,35 +371,18 @@ long unix_rename (MAILSTREAM *stream,char *old,char *newname)
 
 MAILSTREAM *unix_open (MAILSTREAM *stream)
 {
-  long i;
   int fd;
   char tmp[MAILTMPLEN];
 				/* return prototype for OP_PROTOTYPE call */
   if (!stream) return &unixproto;
-  if (LOCAL) {			/* close old file if stream being recycled */
-    unix_close (stream,NIL);	/* dump and save the changes */
-    stream->dtb = &unixdriver;/* reattach this driver */
-    mail_free_cache (stream);	/* clean up cache */
-				/* flush user flags */
-    for (i = 0; i < NUSERFLAGS; i++)
-      if (stream->user_flags[i]) fs_give ((void **) &stream->user_flags[i]);
-  }
+  if (stream->local) fatal ("unix recycle stream");
   stream->local = memset (fs_get (sizeof (UNIXLOCAL)),0,sizeof (UNIXLOCAL));
 				/* canonicalize the stream mailbox name */
   dummy_file (tmp,stream->mailbox);
 				/* canonicalize name */
   fs_give ((void **) &stream->mailbox);
+				/* save canonical name */
   stream->mailbox = cpystr (tmp);
-  /* You may wonder why LOCAL->name is needed.  It isn't at all obvious from
-   * the code.  The problem is that when a stream is recycled with another
-   * mailbox of the same type, the driver's close method isn't called because
-   * it could be IMAP and closing then would defeat the entire point of
-   * recycling.  Hence there is code in the file drivers to call the close
-   * method such as what appears above.  The problem is, by this point,
-   * mail_open() has already changed the stream->mailbox name to point to the
-   * new name, and unix_close() needs the old name.
-   */
-  LOCAL->name = cpystr (tmp);	/* local copy for recycle case */
   LOCAL->fd = LOCAL->ld = -1;	/* no file or state locking yet */
   LOCAL->lname = NIL;
   LOCAL->filesize = 0;		/* initialize file information */
@@ -385,7 +393,7 @@ MAILSTREAM *unix_open (MAILSTREAM *stream)
 
   LOCAL->dirty = NIL;		/* no update yet */
   if (!stream->rdonly) {	/* make lock for read/write access */
-    if ((fd = lockname (tmp,LOCAL->name,NIL)) < 0)
+    if ((fd = lockname (tmp,stream->mailbox,NIL)) < 0)
       mm_log ("Can't open mailbox lock, access is readonly",WARN);
 				/* can get the lock? */
     else if (flock (fd,LOCK_EX|LOCK_NB)) {
@@ -401,7 +409,7 @@ MAILSTREAM *unix_open (MAILSTREAM *stream)
 				/* parse mailbox */
   stream->nmsgs = stream->recent = 0;
 				/* will we be able to get write access? */
-  if ((LOCAL->ld >= 0) && access (LOCAL->name,02) && (errno == EACCES)) {
+  if ((LOCAL->ld >= 0) && access (stream->mailbox,02) && (errno == EACCES)) {
     mm_log ("Can't get write access to mailbox, access is readonly",WARN);
     flock (LOCAL->ld,LOCK_UN);	/* release the lock */
     close (LOCAL->ld);		/* close the lock file */
@@ -621,7 +629,7 @@ long unix_ping (MAILSTREAM *stream)
     }
     else {			/* get current mailbox size */
       if (LOCAL->fd >= 0) fstat (LOCAL->fd,&sbuf);
-      else stat (LOCAL->name,&sbuf);
+      else stat (stream->mailbox,&sbuf);
 				/* parse if mailbox changed */
       if ((sbuf.st_size != LOCAL->filesize) &&
 	  unix_parse (stream,lock,LOCK_SH)) {
@@ -715,12 +723,12 @@ long unix_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
     break;
   case EINVAL:
     if (pc) return (*pc) (stream,sequence,mailbox,options);
-    sprintf (LOCAL->buf,"Invalid UNIX-format mailbox name: %s",mailbox);
+    sprintf (LOCAL->buf,"Invalid UNIX-format mailbox name: %.80s",mailbox);
     mm_log (LOCAL->buf,ERROR);
     return NIL;
   default:
     if (pc) return (*pc) (stream,sequence,mailbox,options);
-    sprintf (LOCAL->buf,"Not a UNIX-format mailbox: %s",mailbox);
+    sprintf (LOCAL->buf,"Not a UNIX-format mailbox: %.80s",mailbox);
     mm_log (LOCAL->buf,ERROR);
     return NIL;
   }
@@ -921,7 +929,6 @@ long unix_append_putc (int fd,char *s,long *i,char c)
 void unix_abort (MAILSTREAM *stream)
 {
   if (LOCAL) {			/* only if a file is open */
-    if (LOCAL->name) fs_give ((void **) &LOCAL->name);
     if (LOCAL->fd >= 0) close (LOCAL->fd);
     if (LOCAL->ld >= 0) {	/* have a mailbox lock? */
       flock (LOCAL->ld,LOCK_UN);/* yes, release the lock */
@@ -959,11 +966,8 @@ int unix_lock (char *file,int flags,int mode,char *lock,int op)
     if ((ld = open(lock,O_BINARY|O_WRONLY|O_CREAT|O_EXCL,S_IREAD|S_IWRITE))>=0)
       close (ld);		/* got it, close the lock file! */
     else if (errno != EEXIST) {	/* miscellaneous error */
-      sprintf (tmp,"Mailbox vulnerable - error creating %s: %s",
-	       lock,strerror (errno));
-      mm_log (tmp,WARN);	/* this is probably not good */
-      *lock = '\0';		/* cancel lock file */
-      i = 0;			/* time out the lock */
+      sprintf (tmp,"Error creating %.80s: %s",lock,strerror (errno));
+      if (!(i%15)) mm_log (tmp,WARN);
     }
 				/* lock exists, still active? */
     else if (!stat (lock,&sb) && (t > sb.st_ctime + LOCKTIMEOUT * 60) &&
@@ -971,7 +975,7 @@ int unix_lock (char *file,int flags,int mode,char *lock,int op)
       close (ld);		/* got timed-out lock file */
     else {			/* active lock, try again */
       if (!(i%15)) {
-	sprintf (tmp,"Mailbox %s is locked, will override in %d seconds...",
+	sprintf (tmp,"Mailbox %.80s is locked, will override in %d seconds...",
 		 file,i);
 	mm_log (tmp,WARN);
       }
@@ -1004,7 +1008,7 @@ void unix_unlock (int fd,MAILSTREAM *stream,char *lock)
 				/* set times to now */
     times.modtime = times.actime = time (0);
 				/* set the times, note change */
-    if (!utime (LOCAL->name,&times)) LOCAL->filetime = times.modtime;
+    if (!utime (stream->mailbox,&times)) LOCAL->filetime = times.modtime;
   }
   flock (fd,LOCK_UN);		/* release flock'ers */
   if (!stream) close (fd);	/* close the file if no stream */
@@ -1022,11 +1026,11 @@ void unix_unlock (int fd,MAILSTREAM *stream,char *lock)
 int unix_parse (MAILSTREAM *stream,char *lock,int op)
 {
   int zn;
-  unsigned long i,j,k;
+  unsigned long i,j,k,m;
   char c,*s,*t,*u,tmp[MAILTMPLEN],date[30];
   int ti = 0,pseudoseen = NIL;
-  unsigned long prevuid = 0;
   unsigned long nmsgs = stream->nmsgs;
+  unsigned long prevuid = nmsgs ? mail_elt (stream,nmsgs)->private.uid : 0;
   unsigned long recent = stream->recent;
   unsigned long oldnmsgs = stream->nmsgs;
   short silent = stream->silent;
@@ -1039,7 +1043,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 				/* toss out previous descriptor */
   if (LOCAL->fd >= 0) close (LOCAL->fd);
   mm_critical (stream);		/* open and lock mailbox (shared OK) */
-  if ((LOCAL->fd = unix_lock (LOCAL->name,
+  if ((LOCAL->fd = unix_lock (stream->mailbox,
 			      O_BINARY + ((LOCAL->ld >= 0) ? O_RDWR:O_RDONLY),
 			      NIL,lock,op)) < 0) {
     sprintf (tmp,"Mailbox open failed, aborted: %s",strerror (errno));
@@ -1114,7 +1118,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 	}
 	else date[18] = date[19] = '0';
 				/* yy -- advance over timezone if necessary */
-	if (zn == ti) ti += (((t[zn] == '+') || (t[zn] == '-')) ? 6 : 4);
+	if (zn == ti) ti += (((t[zn+1] == '+') || (t[zn+1] == '-')) ? 6 : 4);
 	date[7] = t[ti + 1]; date[8] = t[ti + 2];
 	date[9] = t[ti + 3]; date[10] = t[ti + 4];
 				/* zzz */
@@ -1150,7 +1154,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 				/* find end of keyword */
 		  if (!(u = strpbrk (s," \015\012"))) u = s + strlen (s);
 				/* got a keyword? */
-		  if (k = (u - s)) {
+		  if ((k = (u - s)) && (k < MAILTMPLEN)) {
 				/* copy keyword */
 		    strncpy (uf,s,k);
 				/* make sure tied off */
@@ -1243,15 +1247,18 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 				/* make sure not duplicated */
 		  if (elt->private.uid)
 		    sprintf (tmp,"Message %lu UID %lu already has UID %lu",
-			     elt->msgno - 1,j,elt->private.uid);
+			     pseudoseen ? elt->msgno - 1 : elt->msgno,
+			     j,elt->private.uid);
 				/* make sure UID doesn't go backwards */
 		  else if (j <= prevuid)
 		    sprintf (tmp,"Message %lu UID %lu less than %lu",
-			     elt->msgno - 1,j,prevuid + 1);
+			     pseudoseen ? elt->msgno - 1 : elt->msgno,
+			     j,prevuid + 1);
 				/* or skip by mailbox's recorded last */
 		  else if (j > stream->uid_last)
 		    sprintf (tmp,"Message %lu UID %lu greater than last %lu",
-			     elt->msgno - 1,j,stream->uid_last);
+			     pseudoseen ? elt->msgno - 1 : elt->msgno,
+			     j,stream->uid_last);
 		  else {	/* normal UID case */
 		    prevuid = elt->private.uid = j;
 		    break;		/* exit this cruft */
@@ -1299,7 +1306,8 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 	    }
 				/* otherwise fall into default case */
 	  default:		/* ordinary header line */
-	    elt->rfc822_size += i + 1;
+	    elt->rfc822_size += i + 
+	      (((i < 2) || (s[i - 2] != '\015')) ? 1 : 0);
 	    break;
 	  }
 	} while (i && (*t != '\015') && (*t != '\012'));
@@ -1312,7 +1320,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 	  (elt->private.msg.text.offset =
 	   (LOCAL->filesize + GETPOS (&bs)) - elt->private.special.offset) -
 	     elt->private.special.text.size;
-	k = 0;			/* no previous line size yet */
+	k = m = 0;		/* no previous line size yet */
 				/* note current position */
 	j = LOCAL->filesize + GETPOS (&bs);
 	if (i) do {		/* look for next message */
@@ -1321,7 +1329,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 	    VALID (s,t,ti,zn);	/* yes, parse line */
 	    if (!ti) {		/* not a header line, add it to message */
 	      elt->rfc822_size += 
-		k = i + (((i > 1) && s[i - 2] == '\015') ? 0 : 1);
+		k = i + (m = (((i < 2) || s[i - 2] != '\015') ? 1 : 0));
 				/* update current position */
 	      j = LOCAL->filesize + GETPOS (&bs);
 	    }
@@ -1330,7 +1338,7 @@ int unix_parse (MAILSTREAM *stream,char *lock,int op)
 	elt->private.msg.text.text.size = j -
 	  (elt->private.special.offset + elt->private.msg.text.offset);
 	if (k == 2) {		/* last line was blank? */
-	  elt->private.msg.text.text.size -= 2;
+	  elt->private.msg.text.text.size -= (m ? 1 : 2);
 	  elt->rfc822_size -= 2;
 	}
       } while (i);		/* until end of buffer */
@@ -1507,6 +1515,7 @@ long unix_rewrite (MAILSTREAM *stream,unsigned long *nexp)
 {
   unsigned long i,j;
   int e,retry;
+  struct stat sbuf;
   FILE *f;
   MESSAGECACHE *elt;
   unsigned long recent = stream->recent;
@@ -1526,8 +1535,17 @@ long unix_rewrite (MAILSTREAM *stream,unsigned long *nexp)
   else for (i = 1; i <= stream->nmsgs; i++)
     if (!unix_write_message (f,stream,mail_elt (stream,i),&size))
       return unix_punt_scratch (f);
-				/* writing remaining data */
-  if (fflush (f)) return unix_punt_scratch (f);
+				/* write remaining data */
+  if (fflush (f) || fstat (fileno (f),&sbuf)) return unix_punt_scratch (f);
+				/* make damn sure stdio isn't lying */
+  if (size != (unsigned long) sbuf.st_size) {
+    char tmp[MAILTMPLEN];
+    sprintf (tmp,"Checkpoint file size mismatch (%lu != %lu)",size,
+	     sbuf.st_size);
+    mm_log (tmp,ERROR);
+    fclose (f);			/* flush the output file */
+    return NIL;
+  }
 				/* does the mailbox need to grow? */
   if (size > (unsigned long) LOCAL->filesize) {
 				/* am I paranoid or what? */
@@ -1577,7 +1595,7 @@ long unix_rewrite (MAILSTREAM *stream,unsigned long *nexp)
     retry = NIL;		/* no need to retry yet */
     fseek (f,0,L_SET);		/* rewind files */
     lseek (LOCAL->fd,0,L_SET);
-    for (i = size; i; i -= j) {
+    for (i = size; i; i -= j)
       if (!((j = fread (LOCAL->buf,1,min ((long) CHUNK,i),f)) &&
 	    (write (LOCAL->fd,LOCAL->buf,j) >= 0))) {
 	sprintf (LOCAL->buf,"Mailbox rewrite error: %s",strerror (e = errno));
@@ -1586,7 +1604,6 @@ long unix_rewrite (MAILSTREAM *stream,unsigned long *nexp)
 	retry = T;		/* must retry */
 	break;
       }
-    }
   } while (retry);		/* in case need to retry */
   fclose (f);			/* finished with scratch file */
 				/* make sure tied off */
